#include "game.h"
#include "io.h"
#include "menu.h"
#include "time.h"
#include "solve.h"
#include "file.h"

static char offsetX; // space on left side of the screen
static char offsetY; // space on upper side of the screen

static char sizeCell; // width and height of one cell
static char sizeField; // size of the whole field in rows/columns
static char sqrtSizeField; // sqrt(sizeField), needed by some functions

static char font; // small (F_4x6), normal (F_6x8), big (F_8x10)
static char fontOffsetX; // space on left side of the cell
static char fontOffsetY; // space on upper side of the cell

static char selectedX; // coordinates of the selected cell
static char selectedY;

static struct Statistics statistics;

// the complete field including given and solved values, storing one instead of
// two values per byte/char to keep operations simple and fast instead of
// saving memory. set to maximum size (16x16)
Field field;
// the given field
Field givenField;
// the solved field
Field solvedField;

void initialize(char boxes)
{
    statistics.checks = statistics.validations = statistics.hints = statistics.solves = statistics.saves = 0;

    // initialization of the fields
    memset(field, -1, sizeof(Field));
    memset(solvedField, -1, sizeof(Field));
    memset(givenField, 0, sizeof(Field));

    sqrtSizeField = boxes;
    sizeField = boxes * boxes;

    // optimal (biggest possible) size
    sizeCell = 127 / sizeField;

    // center
    offsetY = (127 - sizeCell * sizeField) / 2;
    offsetX = (239 - sizeCell * sizeField) / 2;

    if (sizeCell > 10) // big boxes -> big font
    {
        font = F_8x10;
        fontOffsetX = 3;
        fontOffsetY = 4;
    }
    else if(sizeCell > 8) // medium boxes -> medium font
    {
        font = F_6x8;
        fontOffsetX = 2;
        fontOffsetY = 3;
    }
    else // small boxes -> small font
    {
        font = F_4x6;
        fontOffsetX = 1;
        fontOffsetY = 2;
    }

}

void drawGame()
{
    int i, x, y;

    clearScreen();

    for (i = 0; i <= sizeField; ++i) // normal horizontal lines
        drawLine(
            offsetX,
            offsetY + sizeCell * i,
            offsetX + sizeCell * sizeField, 
            offsetY + sizeCell * i, COLOR_LIGHTGRAY);

    for (i = 0; i <= sizeField; ++i) // normal vertical lines
        drawLine(
            offsetX + sizeCell * i,
            offsetY,
            offsetX + sizeCell * i, 
            offsetY + sizeCell * sizeField, COLOR_LIGHTGRAY);

    // horizontal block lines
    for (i = 0; i <= sizeField; i += sqrtSizeField)
        drawLine(
            offsetX,
            offsetY + sizeCell * i,
            offsetX + sizeCell * sizeField,
            offsetY + sizeCell * i,
            COLOR_BLACK);

    // vertical block lines 
    for (i = 0; i <= sizeField; i += sqrtSizeField)
        drawLine(
            offsetX + sizeCell * i,
            offsetY,
            offsetX + sizeCell * i,
            offsetY + sizeCell * sizeField,
            COLOR_BLACK);

    // fill in the numbers
    for (y = 0; y < sizeField; ++y)
    {
        for (x = 0; x < sizeField; ++x)
        {
            writeCell(x, y);
        }
    }
}

void writeCell(int x, int y)
{
    char value[2] = " \0"; // doesn't \0 else (end string)

    if (field[y][x] < 0) // empty cell
        value[0] = ' ';
    else // normal, filled cell
        value[0] = VARIABLES[(int)field[y][x]];

    write(
        offsetX + x * sizeCell + sizeCell / 2 - fontOffsetX,
        offsetY + y * sizeCell + sizeCell / 2 - fontOffsetY,
        value,
        A_XOR,
        font);
}

void redrawCell(int x, int y)
{
    fillRect( // clear the cell
        offsetX + sizeCell * x + 1,
        offsetY + sizeCell * y + 1,
        offsetX + sizeCell * (x + 1) - 1,
        offsetY + sizeCell * (y + 1) - 1,
        COLOR_WHITE);

    writeCell(x, y);

    // highlight cell if it's selected
    if (x == selectedX && y == selectedY) invertCell(x, y);
}

void invertCell(char x, char y)
{
    invertRect(
        offsetX + sizeCell * x + 1,
        offsetY + sizeCell * y + 1,
        offsetX + sizeCell * (x + 1) - 1,
        offsetY + sizeCell * (y + 1) - 1);
}

void highlightCell(char x, char y)
{
    partInvertRect(
        offsetX + sizeCell * x + 1,
        offsetY + sizeCell * y + 1,
        offsetX + sizeCell * (x + 1) - 1,
        offsetY + sizeCell * (y + 1) - 1,
        COLOR_LIGHTGRAY);
}

void highlightGiven()
{
    int x, y;
    for (y = 0; y < 9; ++y)
        for (x = 0; x < 9; ++x)
            if (givenField[y][x])
                highlightCell(x, y);
}

void selectCell(char x, char y)
{
    if (selectedX >= 0 && selectedY >= 0) // ignore invalid values
        invertCell(selectedX, selectedY);
    invertCell(selectedX = x, selectedY = y);
}

char changeCell(int x, int y, short value)
{
    if (givenField[y][x])
        return FALSE; // not allowed to change

    field[y][x] = value;
    return TRUE;
}

void playGame(char hasMenu, char showTime)
{
    short input; // pressed character
    char doHighlightGiven = FALSE; // whether to highlight given cells

    int hintedX = -1; // reset the cell given by a hint
    int hintedY = -1;

    selectedX = -1; // reset selections
    selectedY = -1;

    selectCell(0, 0); // start at corner top left

    resetTimer();

    // user interaction
    while (TRUE)
    {
        if (hasMenu) startTimer();

        if (checkFullField())
        {
            stopTimer(); // counting the time is pretty useless now

            ngetchx(); // give some time to admire

            if (checkField())
            {
                popup("Sudoku solved");
            }
            else
            {
                popup("Wrong solution");
            }
            printStatistics();
            return;
        }

        while (hasMenu && showTime && !kbhit())
            printTime();

        input = ngetchx(); // get the user input (blocking)

        switch (input)
        {
        // arrows and vim-like control
        case 'j': case KEY_DOWN:
            if (selectedY < sizeField - 1)
                selectCell(selectedX, selectedY + 1);
            else
                selectCell(selectedX, 0); // jump to bottom
            break;
        case 'k': case KEY_UP:
            if (selectedY > 0)
                selectCell(selectedX, selectedY - 1);
            else
                selectCell(selectedX, sizeField - 1); // jump to top
            break;
        case 'h': case KEY_LEFT:
            if (selectedX > 0)
                selectCell(selectedX - 1, selectedY);
            else
                selectCell(sizeField - 1, selectedY); // jump to right
            break;
        case 'l': case KEY_RIGHT:
            if (selectedX < sizeField - 1)
                selectCell(selectedX + 1, selectedY);
            else
                selectCell(0, selectedY); // jump to left
            break;
        case KEY_ESC: // return (entering given numbers) or show menu
            stopTimer();

            if (!hasMenu) return; // no menu

            switch (ingameMenu())
            {
            case 0: // Continue
                break;
            case 1: // Check
                ++statistics.checks;
                if (checkField())
                {
                    popup("Checking succeeded");
                }
                break;
            case 2: // Validate
                ++statistics.validations;
                if (wrappedSolve())
                {
                    popup("Validation succeeded");
                }
                else
                {
                    popup("Validation failed");
                }
                break;
            case 3: // Hint
                if (checkFullField())
                {
                    popup("The field is already full");
                    break;
                }

                ++statistics.hints;

                if (!wrappedSolve())
                    break;

                // enter a random number from the solved field into the normal
                // field
                randomize();

                do
                {
                    hintedX = random(sizeField);
                    hintedY = random(sizeField);
                }
                while (field[hintedY][hintedX] >= 0);

                changeCell(hintedX, hintedY, solvedField[hintedY][hintedX]);
                redrawCell(hintedX, hintedY);

                break;
            case 4: // Solve
                ++statistics.solves;
                if (!wrappedSolve())
                    break;

                fillField(solvedField, 0); // make the solution visible
                doHighlightGiven = TRUE; // highlight the field from the start

                break;
            case 5: // Save
                ++statistics.saves;
                if (!saveSudoku(field, sqrtSizeField))
                {
                    popup("Saving failed");
                }
                else
                {
                    popup("Saving succeeded");
                }
                break;
            case 6: // Exit
                printStatistics();
                return;
            }

            drawGame();

            if (doHighlightGiven)
                highlightGiven();

            // highlight the hint (given cell) and wait for a key press
            if (hintedX >= 0 && hintedY >= 0)
            {
                highlightCell(hintedX, hintedY);
                ngetchx();
                highlightCell(hintedX, hintedY);

                hintedX = -1;
                hintedY = -1;
            }

            invertCell(selectedX, selectedY); // show cursor

            break;
        case ' ': case KEY_ENTER: // clear cell
            if (changeCell(selectedX, selectedY, -1))
                redrawCell(selectedX, selectedY);
            break;
        // cell filling
        case '1': case '2': case '3': case '4': case '5': 
        case '6': case '7': case '8': case '9': case '0':
            if (sizeField < 10) // input from 1 to 9 (at max)
            {
                if (input >= '1' && input <= '0' + sizeField)
                {
                    if (changeCell(selectedX, selectedY, input - '0'))
                        redrawCell(selectedX, selectedY);
                }
            }
            else // number area of big fields (0 to 9)
            {
                if (changeCell(selectedX, selectedY, input - '0'))
                    redrawCell(selectedX, selectedY);
            }
            break;
        case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
            // letter area of big fields (a to f)
            if (input >= 'a' && input <= 'a' + sizeField - 11)
            {
                if (changeCell(selectedX, selectedY, input - 'a' + 10))
                    redrawCell(selectedX, selectedY);
            }
            break;
        }
    }
}

char checkField()
{
    if (doCheckField())
        return TRUE;
    popup("Validation failed");
    return FALSE;
}

char doCheckField()
{
    int i, j, k, l, m; // helpers
    char found[2]; // whether there already was a cell with the current value

    for (i = 0; i < sizeField; ++i)
    {
        for (j = 0; j < sizeField; ++j)
        {
            found[0] = FALSE;
            found[1] = FALSE;
            for (k = 0; k < sizeField; ++k)
            {
                if (field[i][k] == j) // validate rows
                {
                    if (found[0]) return FALSE; // duplicate
                    found[0] = TRUE; // appears first time
                }
                if (field[k][i] == j) // validate columns
                {
                    if (found[1]) return FALSE;
                    found[1] = TRUE;
                }
            }
        }
    }

    // validate boxes
    for (i = 0; i < sqrtSizeField; ++i) // "box-column"
    {
        for (j = 0; j < sqrtSizeField; ++j) // "box-row"
        {
            for (k = 0; k < sizeField; ++k) // value
            {
                found[0] = FALSE; // reusing
                for (l = 0; l < sqrtSizeField; ++l) // column
                {
                    for (m = 0; m < sqrtSizeField; ++m) // row
                    {
                        if (field[i * sqrtSizeField + l]
                                 [j * sqrtSizeField + m] == k)
                        {
                            if (found[0]) return FALSE; // duplicate
                            found[0] = TRUE; // appears first time
                        }
                    }
                }
            }
        }
    }
    return TRUE;
}

char checkFullField()
{
    int x, y;
    for (y = 0; y < sizeField; ++y) // column
        for (x = 0; x < sizeField; ++x) // row
            if (field[y][x] < 0)
                return FALSE;
    return TRUE;
}

char validateSolvedField()
{
    int x, y;

    if (checkFullField()) return FALSE;

    for (y = 0; y < sizeField; ++y) // column
        for (x = 0; x < sizeField; ++x) // row
        {
            if (
                (field[y][x] >= 0 && field[y][x] != solvedField[y][x]) ||
                solvedField[y][x] < 0 // empty solution?
            )
                return FALSE;
        }
    return TRUE;

}

void enterField()
{
    playGame(HAS_NO_MENU, DO_NOT_SHOW_TIME);
    invertCell(selectedX, selectedY);
    setGivenField();
}

void fillField(Field newField, char givenField)
{
    memcpy(field, newField, sizeof(Field));
    if (givenField) setGivenField();
}

void setGivenField()
{
    int x, y;
    for (y = 0; y < sizeField; ++y)
        for (x = 0; x < sizeField; ++x)
            if (field[y][x] < 0)
                givenField[y][x] = 0;
            else
                givenField[y][x] = 1;
}

char wrappedSolve()
{
    if (sqrtSizeField != 3) // only implemented for 9x9
    {
        popup("Solving and Hints only work for 9x9");
        return FALSE;
    }

    if (!checkField()) return FALSE; // impossible on invalid fields

    if (!validateSolvedField()) // only do if it not done before
    {
        message("SOLVING");

        solve(field, solvedField);

        clearScreen();
    }
    return TRUE;
}

void printStatistics()
{
    char s[30];
    char t[250];
    getTime(s);
    sprintf(t, "\
Time played: %14s\n\
Checks:      %14d\n\
Validates:   %14d\n\
Hints:       %14d\n\
Solves:      %14d\n\
Saves:       %14d", s, statistics.checks, statistics.validations, statistics.hints, statistics.solves, statistics.saves);
    popup(t);
}
