#ifndef GAME_H
#define GAME_H


// INCLUDES

#include "sudoku.h"


// FUNCTIONS

/**
 * Initializes all variables and calculate all sizes that are neccessary for
 * the other functions. Has to be run first and again for every new game.
 *
 * @param boxes The number of boxes (square root of rows/columns)
 */
void initialize(char boxes);

/**
 * Draw the sudoku field (gray lines for normal, black lines for blocks) and
 * fill it with the available numbers.
 */
void drawGame();

/**
 * Write the number of a cell.
 *
 * @param x The x-coordinate of the cell
 * @param y The y-coordinate of the cell
 */
void writeCell(int x, int y);

/**
 * Draw a cell again and write its number.
 *
 * @param x The x-coordinate of the cell
 * @param y The y-coordinate of the cell
 */
void redrawCell(int x, int y);

/**
 * Invert the colors of a cell.
 *
 * @param x The x-coordinate of the cell
 * @param y The y-coordinate of the cell
 */
void invertCell(char x, char y);

/**
 * Highlight a cell (invert only the light pane)
 *
 * @param x The x-coordinate of the cell
 * @param y The y-coordinate of the cell
 */
void highlightCell(char x, char y);

/**
 * Highlight all cells that were given from the start.
 *
 * @param x The x-coordinate of the cell
 * @param y The y-coordinate of the cell
 */
void highlightGiven();

/**
 * Select a cell and deselect the one selected up to now.
 *
 * @param x The x-coordinate of new the cell
 * @param y The y-coordinate of new the cell
 */
void selectCell(char x, char y);

/**
 * Change the value of a cell.
 *
 * @param x The x-coordinate of new the cell
 * @param y The y-coordinate of new the cell
 */
char changeCell(int x, int y, short input);

/**
 * The main function reacting to user input.
 *
 * @param hasMenu Whether the game has a menu (1) or quits on Escape (0)
 * @param showTime Whether to show the time
 */
void playGame(char hasMenu, char showTime);

/** 
 * Check whether the current field is valid. (no duplicate values in rows,
 * columns and boxes) and output a message if the validation failed.
 */
char checkField();

/**
 * Does the work for checkField()
 */
char doCheckField();

/**
 * Check whether the field is completely filled (no empty (set to -1) cells)
 */
char checkFullField();

/**
 * Check whether the solved field is still coincident with the entered field.
 * This is impossible on real sudokus (only one possible solution), but we
 * might never know what the user enters.
 */
char validateSolvedField();

/**
 * Let the user enter the sudoku by his own (for example from a magazine)
 * instead of generating them internally.
 */
void enterField();

/**
 * Fill the sudoku with a new field but don't draw it. Also set the status for
 * all the filled values.
 *
 * @param newField The field to fill in
 * @param givenField Whether the field is filled with given numbers
 */
void fillField(Field newField, char givenField);

/**
 * Set all filled values (not -1) to given ones.
 */
void setGivenField();

/**
 * Generates a field in the correct size and the set difficulty.
 *
 * @param difficulty The difficulty of the field to generate, see difficulty
 * from sudoku.h
 */
//FieldPointer generateField(char difficulty);

/**
 * Solve the field while printing an info message.
 */
char wrappedSolve();

/**
 * Print the statistics (time played, number of checks, validates, hints, solves, saves)
 */
void printStatistics();

#endif // GAME_H
